import requests
import os
import json
import tempfile

def _load_spec_from_uri(uri: str) -> dict:
    # Convert github blob URL to raw URL
    if (uri.startswith('https://github.com/')) and ('/blob/' in uri):
        raw_url = uri.replace('github.com', 'raw.githubusercontent.com').replace('/blob/', '/') + f'?cachebust={os.urandom(16).hex()}'
    else:
        raw_url = uri

    if raw_url.startswith('file://'):
        # Read the content from a local file
        with open(raw_url[len('file://'):], 'r', encoding='utf-8') as file:
            content = file.read()
    else:
        # Download the content
        response = requests.get(raw_url, timeout=60)
        response.raise_for_status()
        content = response.text

    # Save to a temporary file
    with tempfile.NamedTemporaryFile(delete=False, mode='w', suffix='.json') as temp_file:
        temp_file.write(content)
        temp_file_path = temp_file.name

    # Read the JSON content from the file
    with open(temp_file_path, 'r', encoding='utf-8') as file:
        data = json.load(file)

    # Clean up the temporary file
    os.remove(temp_file_path)

    return data



from typing import List

from ..common.dendro_types import DendroJobRequiredResources


# You must first setup the AWS credentials
# You can do this in multiple ways like using the aws configure command
# or by setting environment variables (AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY, etc.).

class JobDefinitionException(Exception):
    pass

def _run_job_in_aws_batch(
    *,
    job_id: str,
    job_private_key: str,
    app_name: str,
    container: str, # for verifying consistent with job definition
    command: str,
    required_resources: DendroJobRequiredResources
):
    import boto3

    client = boto3.client(
        'batch'
    )

    stack_id = 'DendroBatchStack'
    aws_batch_job_definition = f'{app_name}-dendro-jd'
    if required_resources.numGpus > 0:
        aws_batch_job_queue = f'{stack_id}-job-queue-gpu'
    else:
        aws_batch_job_queue = f'{stack_id}-job-queue-cpu'
    job_def_resp = client.describe_job_definitions(jobDefinitionName=aws_batch_job_definition)
    job_defs = job_def_resp['jobDefinitions']
    if len(job_defs) == 0:
        raise JobDefinitionException(f'Job definition not found: {aws_batch_job_definition}')
    job_def = job_defs[0]
    job_def_container = job_def['containerProperties']['image']
    if job_def_container != container:
        raise JobDefinitionException(f'Job definition container does not match: {job_def_container} != {container}')
    # job_def_command = job_def['containerProperties']['command']
    # if not _command_matches(job_def_command, command):
    #     warn(f'Warning: Job definition command does not match: {job_def_command} != {command}')
    #     # raise JobDefinitionException(f'Job definition command does not match: {job_def_command} != {command}')

    job_name = f'dendro-job-{job_id}'

    env_vars = {
        'JOB_ID': job_id,
        'JOB_PRIVATE_KEY': job_private_key,
        'APP_EXECUTABLE': command,
        'DENDRO_JOB_WORKING_DIR': f'/tmp/dendro-jobs/{job_id}',
        'DENDRO_JOB_CLEANUP_DIR': f'/tmp/dendro-jobs/{job_id}'
    }

    if required_resources.timeSec is not None:
        env_vars['JOB_TIMEOUT_SEC'] = str(int(required_resources.timeSec))

    # Not doing this any more -- instead we are setting a custom backend for kachery uploads
    # from ._start_job import _get_kachery_cloud_credentials # avoid circular import
    # kachery_cloud_client_id, kachery_cloud_private_key = _get_kachery_cloud_credentials()
    # if kachery_cloud_client_id is not None:
    #     env_vars['KACHERY_CLOUD_CLIENT_ID'] = kachery_cloud_client_id
    #     if not kachery_cloud_private_key:
    #         raise KeyError('kachery_cloud_private key is not set even though kachery_cloud_client_id is set')
    #     env_vars['KACHERY_CLOUD_PRIVATE_KEY'] = kachery_cloud_private_key

    resource_requirements_aws = [
        {
            'type': 'VCPU',
            'value': str(required_resources.numCpus)
        },
        {
            'type': 'MEMORY',
            # Luiz: "we should define the max RAM always to be just slightly lower than the ram available,
            # e.g. 63000 instead of 65536 in a ~64 Gb ram instance. This is because the actual ram available
            # in the machine is usually slightly less than the number announced, and the Compute Environment
            # will get stuck trying to find an instance with enough ram and never finding one..."
            'value': str(int(required_resources.memoryGb * 1024 * 0.9))
        }
    ]
    if required_resources.numGpus > 0:
        resource_requirements_aws.append({
            'type': 'GPU',
            'value': str(required_resources.numGpus)
        })
        env_vars['NVIDIA_DRIVER_CAPABILITIES'] = 'all'
        env_vars['NVIDIA_REQUIRE_CUDA'] = 'cuda>=11.0'

    response = client.submit_job(
        jobName=job_name,
        jobQueue=aws_batch_job_queue,
        jobDefinition=aws_batch_job_definition,
        containerOverrides={
            'environment': [
                {
                    'name': k,
                    'value': v
                }
                for k, v in env_vars.items()
            ],
            'resourceRequirements': resource_requirements_aws
        },
        timeout={
            # Add a buffer because dendro will be the one to actually terminate the job and do the cleanup
            # based on the required_resources.timeSec parameter that the job also has access to.
            # We assume that the startup and cleanup operations take less than the buffer time.
            'attemptDurationSeconds': int(required_resources.timeSec + 60 * 10)
        }
    )

    batch_job_id = response['jobId']
    print(f'AWS Batch job submitted: {job_id} {batch_job_id}')

def _command_matches(cmd1: List[str], cmd2: str) -> bool:
    return ' '.join(cmd1) == cmd2



import os
import sys
import subprocess
from typing import Union, Dict, Any, Literal

from ..sdk.App import App
from ..common._api_request import _processor_put_api_request
from ..common.dendro_types import DendroJobRequiredResources
from ._run_job_in_aws_batch import _run_job_in_aws_batch
from ..mock import using_mock


class JobException(Exception):
    pass

def _set_job_status_to_starting(*,
    job_id: str,
    job_private_key: str
):
    url_path = f'/api/processor/jobs/{job_id}/status'
    headers = {
        'job-private-key': job_private_key
    }
    data: Dict[str, Any] = {
        'status': 'starting'
    }
    if os.environ.get('DENDRO_FORCE_STATUS_UPDATES', None) == '1':
        data['force_update'] = True # pragma: no cover
    resp = _processor_put_api_request(
        url_path=url_path,
        headers=headers,
        data=data
    )
    if not resp['success']:
        raise JobException(f'Error setting job status to starting: {resp["error"]}') # pragma: no cover

def _start_job(*,
    job_id: str,
    job_private_key: str,
    processor_name: str,
    run_method: Literal['local', 'slurm', 'aws_batch'],
    app: App,
    run_process: bool = True,
    return_shell_command: bool = False,
    required_resources: DendroJobRequiredResources
):
    assert not (return_shell_command and run_process), 'Cannot set both run_process and return_shell_command to True'
    assert return_shell_command or run_process, 'Cannot set both run_process and return_shell_command to False'

    _set_job_status_to_starting(
        job_id=job_id,
        job_private_key=job_private_key
    )
    assert hasattr(app, '_app_executable'), 'App does not have an executable path'
    app_executable: Union[str, None] = app._app_executable
    app_image: Union[str, None] = app._app_image

    # default for app_executable
    if app_executable is None:
        assert app_image, 'You must set app_executable if app_image is not set'
        app_executable = '/app/main.py' # the default

    if run_method == 'slurm':
        assert not run_process, 'Not expecting to see run_process here'

    if run_method == 'aws_batch':
        assert not return_shell_command, 'Cannot return shell command for AWS Batch job'
        assert app_image, 'aws_batch_job_queue is set but app_image is not set'
        print(f'Running job in AWS Batch: {job_id} {processor_name}')
        try:
            _run_job_in_aws_batch(
                job_id=job_id,
                job_private_key=job_private_key,
                app_name=app._name,
                container=app_image, # for verifying consistent with job definition
                command=app_executable,
                required_resources=required_resources
            )
        except Exception as e:
            raise JobException(f'Error running job in AWS Batch: {e}') from e
        return ''

    # WARNING!!! The working_dir is going to get cleaned up after the job is finished
    # so it's very important to not set the working directory to a directory that is
    # used for other purposes
    working_dir = os.getcwd() + '/jobs/' + job_id
    os.makedirs(working_dir, exist_ok=True)

    env_vars = {
        'PYTHONUNBUFFERED': '1',
        'JOB_ID': job_id,
        'JOB_PRIVATE_KEY': job_private_key,
        'APP_EXECUTABLE': app_executable,
        'DENDRO_URL': 'https://dendro.vercel.app'
    }

    if required_resources.timeSec is not None:
        env_vars['JOB_TIMEOUT_SEC'] = str(int(required_resources.timeSec))

    # Not doing this any more -- instead we are setting a custom backend for kachery uploads
    # kachery_cloud_client_id, kachery_cloud_private_key = _get_kachery_cloud_credentials()
    # if kachery_cloud_client_id is not None:
    #     env_vars['KACHERY_CLOUD_CLIENT_ID'] = kachery_cloud_client_id
    #     assert kachery_cloud_private_key, 'Unexpected: kachery_cloud_private_key is not set even though kachery_cloud_client_id is set'
    #     env_vars['KACHERY_CLOUD_PRIVATE_KEY'] = kachery_cloud_private_key

    if not app_image:
        # for safety, verify that cleanup directory is as expected
        dendro_job_cleanup_dir = working_dir
        assert dendro_job_cleanup_dir.endswith('/jobs/' + job_id), f'Unexpected dendro_job_cleanup_dir: {dendro_job_cleanup_dir}'
        env_vars['DENDRO_JOB_CLEANUP_DIR'] = dendro_job_cleanup_dir # see the warning above
        return _run_local_job(
            app_executable=app_executable,
            env_vars=env_vars,
            working_dir=working_dir,
            run_process=run_process,
            return_shell_command=return_shell_command
        )

    return _run_container_job(
        app_executable=app_executable,
        app_image=app_image,
        env_vars=env_vars,
        working_dir=working_dir,
        run_process=run_process,
        return_shell_command=return_shell_command,
        num_cpus=required_resources.numCpus,
        use_gpu=required_resources.numGpus > 0
        # don't actually limit the memory, because we don't want the process being harshly terminated - it needs to be able to clean up
    )

# This was the method used previously when we wanted to capture the output of the process and display it to the console
# However, that was problematic, because when this parent closes, we don't want a broken pipe
# prefix = f'{job_id} {processor_name}: '
# t1 = threading.Thread(target=stream_output, args=(process.stdout, prefix))
# t1.start()
# prefix = f'{job_id} {processor_name} ERR: '
# t2 = threading.Thread(target=stream_output, args=(process.stderr, prefix))
# t2.start()

# previously did this (see above)
# def stream_output(pipe, prefix: str):
#     while True:
#         try:
#             line = pipe.readline()
#         except:
#             break
#         if line:
#             print(prefix + line.decode('utf-8'))
#         else:
#             break

def _run_local_job(*,
    app_executable: str,
    env_vars: dict,
    working_dir: str,
    run_process: bool,
    return_shell_command: bool,
):
    if using_mock():
        print('Running job directly for mock testing')

        from ..sdk.App import App

        app_instance: App = _load_app_from_main(app_executable)
        try:
            old_environ = {}
            for k, v in env_vars.items():
                old_environ[k] = os.environ.get(k, None)
                os.environ[k] = v

            # Call the make_spec_file method
            print('MOCK: running app instance')
            app_instance.run()
            print('MOCK: Done running app instance')

            for k, v in old_environ.items():
                if v is None:
                    del os.environ[k]
                else:
                    os.environ[k] = v
        finally:
            if 'main' in sys.modules:
                del sys.modules['main'] # important to do this so that at a later time we can load a different main.py

    if run_process:
        if using_mock():
            return
        print(f'Running: {app_executable}')
        subprocess.Popen(
            [app_executable],
            cwd=working_dir,
            start_new_session=True, # This is important so it keeps running even if the compute resource is stopped
            # Important to set output to devnull so that we don't get a broken pipe error if this parent process is closed
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            env={
                **os.environ,
                **env_vars
            }
        )
        return ''
    elif return_shell_command:
        env_vars_str = ' '.join([f'{k}={v}' for k, v in env_vars.items()])
        return f'cd {working_dir} && {env_vars_str} APP_EXECUTABLE={app_executable} {app_executable}'
    else:
        return ''

def _run_container_job(*,
    app_executable: str,
    app_image: str,
    env_vars: dict,
    working_dir: str,
    run_process: bool,
    return_shell_command: bool,
    num_cpus: Union[int, None],
    use_gpu: bool
):
    container_method = os.environ.get('CONTAINER_METHOD', 'docker')
    if container_method == 'docker':
        tmpdir = working_dir + '/tmp'
        os.makedirs(tmpdir, exist_ok=True)
        os.makedirs(tmpdir + '/working', exist_ok=True)
        cmd2 = [
            'docker', 'run', '-it'
        ]
        cmd2.extend(['-v', f'{tmpdir}:/tmp'])
        env_vars['DENDRO_JOB_CLEANUP_DIR'] = '/tmp/working'
        cmd2.extend(['--workdir', '/tmp/working']) # the working directory will be /tmp/working
        for k, v in env_vars.items():
            cmd2.extend(['-e', f'{k}={v}'])
        # we want kachery temporary files to be stored in the /tmp/.kachery-cloud directory
        cmd2.extend(['-e', 'KACHERY_CLOUD_DIR=/tmp/.kachery-cloud'])
        if num_cpus is not None:
            cmd2.extend(['--cpus', str(num_cpus)])
        if use_gpu:
            cmd2.extend(['--gpus', 'all'])
        cmd2.extend([app_image])
        cmd2.extend([app_executable])
        if run_process:
            print(f'Pulling image {app_image}')
            subprocess.run(['docker', 'pull', app_image])
            print(f'Running: {" ".join(cmd2)}')
            subprocess.Popen(
                cmd2,
                cwd=working_dir,
                start_new_session=True, # This is important so it keeps running even if the compute resource is stopped
                # Important to set output to devnull so that we don't get a broken pipe error if this parent process is closed
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            return ''
        elif return_shell_command:
            return f'cd {working_dir} && {" ".join(cmd2)}'
        else:
            return ''
    elif container_method == 'singularity' or container_method == 'apptainer':
        tmpdir = working_dir + '/tmp' # important to provide a /tmp directory for singularity or apptainer so that it doesn't run out of disk space
        os.makedirs(tmpdir, exist_ok=True)
        os.makedirs(tmpdir + '/working', exist_ok=True)

        # determine the appropriate executable
        if container_method == 'singularity':
            executable = 'singularity'
        elif container_method == 'apptainer':
            executable = 'apptainer'
        else:
            raise JobException(f'Unexpected container method (*): {container_method}')

        cmd2 = [executable, 'exec']
        cmd2.extend(['--bind', f'{tmpdir}:/tmp'])
        # The working directory should be /tmp/working so that if the container wants to write to the working directory, it will not run out of space
        env_vars['DENDRO_JOB_CLEANUP_DIR'] = '/tmp/working'
        cmd2.extend(['--pwd', '/tmp/working'])
        cmd2.extend(['--cleanenv']) # this is important to prevent singularity or apptainer from passing environment variables to the container
        cmd2.extend(['--contain']) # we don't want singularity or apptainer to mount the home or tmp directories of the host
        if use_gpu:
            cmd2.extend(['--nv'])
        for k, v in env_vars.items():
            cmd2.extend(['--env', f'{k}={v}'])
        # we want kachery temporary files to be stored in the /tmp/.kachery-cloud directory
        cmd2.extend(['--env', 'KACHERY_CLOUD_DIR=/tmp/.kachery-cloud'])

        # don't use --cpus for now because we run into cgroups issues and the container fails to start
        # if num_cpus is not None:
        #     cmd2.extend(['--cpus', str(num_cpus)])

        cmd2.extend([f'docker://{app_image}']) # todo: what if it's not a dockerhub image?
        cmd2.extend([app_executable])
        if run_process:
            print(f'Running: {" ".join(cmd2)}')
            subprocess.Popen(
                cmd2,
                cwd=working_dir,
                start_new_session=True, # This is important so it keeps running even if the compute resource is stopped
                # Important to set output to devnull so that we don't get a broken pipe error if this parent process is closed
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            return ''
        elif return_shell_command:
            return f'cd {working_dir} && {" ".join(cmd2)}'
        else:
            return ''
    else:
        raise JobException(f'Unexpected container method: {container_method}')

def _get_kachery_cloud_credentials():
    try:
        from kachery_cloud._client_keys import _get_client_keys_hex
    except ImportError:
        print('Warning: unable to import client keys from kachery_cloud')
        return None, None
    client_id, private_key = _get_client_keys_hex()
    return client_id, private_key

def _load_app_from_main(py_file_path: str) -> App:
    import sys

    # Split the file path into directory and module name (without .py)
    dir_path, file_name = os.path.split(py_file_path)
    module_name = os.path.splitext(file_name)[0]
    if module_name != 'main':
        raise Exception(f'Unexpected module name (expected main): {module_name}')

    # Save the original sys.path
    original_sys_path = [p for p in sys.path] # important to create a copy here

    # Prepend the directory path to sys.path
    sys.path.insert(0, dir_path)

    try:
        import main # type: ignore
        app = main.app
    finally:
        # Restore the original sys.path
        sys.path = original_sys_path

    return app



import hashlib
import uuid
from fastapi import APIRouter, Header
from pydantic import BaseModel
from ..common import api_route_wrapper
from ._authenticate_gui_request import _authenticate_gui_request
from ...clients.db import set_dendro_api_key_for_user


router = APIRouter()

class CreateDendroApiKeyResponse(BaseModel):
    dendroApiKey: str
    success: bool

@router.post("/{user_id}/dendro_api_key")
@api_route_wrapper
async def create_dendro_api_key(user_id, github_access_token: str = Header(...)):
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    # create the api key as the sha1 hash a of uuid
    new_dendro_api_key = hashlib.sha1(str(uuid.uuid4()).encode('utf-8')).hexdigest()

    # set the api key for the user
    await set_dendro_api_key_for_user(user_id, new_dendro_api_key)

    return CreateDendroApiKeyResponse(dendroApiKey=new_dendro_api_key, success=True)



import os
from ..common._api_request import _processor_get_api_request


def _get_upload_url(*, job_id: str, job_private_key: str, output_name: str) -> str:
    """Get a signed upload URL for the output (console or resource log) of a job"""
    url_path = f'/api/processor/jobs/{job_id}/outputs/{output_name}/upload_url'
    headers = {
        'job-private-key': job_private_key
    }
    res = _processor_get_api_request(
        url_path=url_path,
        headers=headers
    )
    return res['uploadUrl']

def _process_is_alive(pid: str) -> bool:
    """
    Check if a process is alive.
    """
    try:
        os.kill(int(pid), 0)
        return True
    except OSError:
        return False



import time
from typing import Union, List, Any, Literal

from ....common.dendro_types import ComputeResourceSpecProcessor, DendroJobInputFile, DendroJobOutputFile, DendroJob, DendroJobInputParameter, DendroJobRequiredResources
from ...clients.db import fetch_project, fetch_file, delete_file, fetch_project_jobs, delete_job, insert_job
from ...core._get_project_role import _check_user_can_edit_project
from ...core._create_random_id import _create_random_id
from ...clients.pubsub import publish_pubsub_message
from .._remove_detached_files_and_jobs import _remove_detached_files_and_jobs
from ...core.settings import get_settings
from ....common.dendro_types import CreateJobRequestInputFile, CreateJobRequestOutputFile, CreateJobRequestInputParameter
from .._create_output_file import _create_output_file

class CreateJobException(Exception):
    pass

async def create_job(*,
    project_id: str,
    processor_name: str,
    input_files_from_request: List[CreateJobRequestInputFile],
    output_files_from_request: List[CreateJobRequestOutputFile],
    input_parameters: List[CreateJobRequestInputParameter],
    processor_spec: ComputeResourceSpecProcessor,
    batch_id: Union[str, None],
    user_id: str,
    dandi_api_key: Union[str, None] = None,
    required_resources: DendroJobRequiredResources,
    run_method: Literal['local', 'aws_batch', 'slurm'],
    pending_approval: bool = False
):
    _check_job_is_consistent_with_processor_spec(
        processor_spec=processor_spec,
        processor_name=processor_name,
        input_files_from_request=input_files_from_request,
        output_files_from_request=output_files_from_request,
        input_parameters=input_parameters,
        required_resources=required_resources
    )

    project = await fetch_project(project_id)
    assert project is not None, f"No project with ID {project_id}"

    _check_user_can_edit_project(project, user_id)

    compute_resource_id = project.computeResourceId if project.computeResourceId else get_settings().DEFAULT_COMPUTE_RESOURCE_ID
    if not compute_resource_id:
        raise KeyError('Project does not have a compute resource ID, and no default VITE_DEFAULT_COMPUTE_RESOURCE_ID is set in the environment.')

    input_files: List[DendroJobInputFile] = [] # {name, fileId, fileName}
    for input_file in input_files_from_request:
        file = await fetch_file(project_id, input_file.fileName)
        if file is None:
            raise CreateJobException(f"Project input file does not exist: {input_file.fileName}")
        input_files.append(
            DendroJobInputFile(
                name=input_file.name,
                fileId=file.fileId,
                fileName=file.fileName,
                isFolder=input_file.isFolder
            )
        )

    job_id = project_id + '.' + _create_random_id(8)
    job_private_key = _create_random_id(32)

    def filter_output_file_name(file_name):
        # replace ${job-id} with the actual job ID
        return file_name.replace('${job-id}', job_id)

    output_files: List[DendroJobOutputFile] = []
    for output_file in output_files_from_request:
        output_files.append(
            DendroJobOutputFile(
                name=output_file.name,
                fileName=filter_output_file_name(output_file.fileName),
                isFolder=output_file.isFolder
            )
        )

    something_was_deleted = False

    # delete any existing output files
    for output_file in output_files:
        existing_file = await fetch_file(project_id, output_file.fileName)
        if existing_file is not None:
            await delete_file(project_id, output_file.fileName)
            something_was_deleted = True

    # delete any jobs that are expected to produce the output files
    # because maybe the output files haven't been created yet, but we still want to delete/cancel them
    all_jobs = await fetch_project_jobs(project_id, include_private_keys=False)

    output_file_names = [x.fileName for x in output_files]
    for job in all_jobs:
        should_delete = False
        for output_file in job.outputFiles:
            if output_file.fileName in output_file_names:
                should_delete = True
        if should_delete:
            await delete_job(job.jobId)
            something_was_deleted = True

    if something_was_deleted:
        await _remove_detached_files_and_jobs(project_id)

    # Create output files in pending state
    output_file_ids: list[str] = []
    for output_file in output_files:
        output_file_id = await _create_output_file(
            file_name=output_file.fileName,
            url='pending',
            project_id=project_id,
            user_id=user_id,
            job_id=job_id,
            is_folder=output_file.isFolder
        )
        output_file_ids.append(output_file_id)

    input_parameters2: List[DendroJobInputParameter] = []
    for input_parameter in input_parameters:
        pp = next((x for x in processor_spec.parameters if x.name == input_parameter.name), None)
        if not pp:
            raise CreateJobException(f"Processor parameter not found: {input_parameter.name}")
        input_parameters2.append(
            DendroJobInputParameter(
                name=input_parameter.name,
                value=input_parameter.value,
                secret=pp.secret
            )
        )

    output_bucket_base_url = get_settings().OUTPUT_BUCKET_BASE_URL
    job = DendroJob(
        jobId=job_id,
        jobPrivateKey=job_private_key,
        projectId=project_id,
        userId=user_id,
        processorName=processor_name,
        inputFiles=input_files,
        inputFileIds=[x.fileId for x in input_files],
        inputParameters=input_parameters2,
        outputFiles=output_files,
        outputFileIds=output_file_ids,
        timestampCreated=time.time(),
        computeResourceId=compute_resource_id,
        status='pending',
        processorSpec=processor_spec,
        batchId=batch_id,
        dandiApiKey=dandi_api_key,
        consoleOutputUrl=f"{output_bucket_base_url}/dendro-outputs/{job_id}/_console_output",
        resourceUtilizationLogUrl=f"{output_bucket_base_url}/dendro-outputs/{job_id}/_resource_utilization_log",
        requiredResources=required_resources,
        runMethod=run_method,
        pendingApproval=pending_approval
    )

    await insert_job(job)

    await publish_pubsub_message(
        channel=job.computeResourceId,
        message={
            'type': 'newPendingJob',
            'projectId': project_id,
            'computeResourceId': compute_resource_id,
            'jobId': job_id
        }
    )

    return job_id

def _check_job_is_consistent_with_processor_spec(
    processor_spec: ComputeResourceSpecProcessor,
    processor_name: str,
    input_files_from_request: List[CreateJobRequestInputFile],
    output_files_from_request: List[CreateJobRequestOutputFile],
    input_parameters: List[CreateJobRequestInputParameter],
    required_resources: DendroJobRequiredResources
):
    # check that the processor name matches
    if processor_spec.name != processor_name:
        raise CreateJobException(f"Processor name mismatch: {processor_spec.name} != {processor_name}")

    # check that the input files are consistent with the spec
    for input_file in input_files_from_request:
        if not input_file.isFolder:
            if input_file.name.endswith(']'):
                base_name = input_file.name.split('[')[0]
                xx = next((x for x in processor_spec.inputs if x.name == base_name), None)
                if not xx:
                    raise CreateJobException(f"Processor input not found: {base_name}")
                if not xx.list:
                    raise CreateJobException(f"Processor input is not a list: {base_name}")
            else:
                xx = next((x for x in processor_spec.inputs if x.name == input_file.name), None)
                if not xx:
                    raise CreateJobException(f"Processor input not found: {input_file.name}")
        else:
            # folder
            if input_file.name.endswith(']'):
                base_name = input_file.name.split('[')[0]
                if not processor_spec.inputFolders:
                    raise CreateJobException(f"Processor input folder not found (*): {base_name}")
                xx = next((x for x in processor_spec.inputFolders if x.name == base_name), None)
                if not xx:
                    raise CreateJobException(f"Processor input folder not found: {base_name}")
                if not xx.list:
                    raise CreateJobException(f"Processor input folder is not a list: {base_name}")
            else:
                if not processor_spec.inputFolders:
                    raise CreateJobException(f"Processor input folder not found: {input_file.name}")
                xx = next((x for x in processor_spec.inputFolders if x.name == input_file.name), None)
                if not xx:
                    raise CreateJobException(f"Processor input folder not found: {input_file.name}")

    # check that all the required inputs are present
    for input in processor_spec.inputs:
        if not input.list:
            if not any(x.name == input.name and not x.isFolder for x in input_files_from_request):
                raise CreateJobException(f"Required input not found: {input.name}")

    # check that all the required input folders are present
    if processor_spec.inputFolders:
        for input in processor_spec.inputFolders:
            if not input.list:
                if not any(x.name == input.name and x.isFolder for x in input_files_from_request):
                    raise CreateJobException(f"Required input folder not found: {input.name}")

    # check that the output files are consistent with the spec
    for output_file in output_files_from_request:
        if not output_file.isFolder:
            xx = next((x for x in processor_spec.outputs if x.name == output_file.name), None)
            if not xx:
                raise CreateJobException(f"Processor output not found: {output_file.name}")
        else:
            # folder
            xx = next((x for x in (processor_spec.outputFolders or []) if x.name == output_file.name), None)
            if not xx:
                raise CreateJobException(f"Processor output folder not found: {output_file.name}")

    # check that the required output files are present
    for output in processor_spec.outputs:
        if not any(x.name == output.name and not x.isFolder for x in output_files_from_request):
            raise CreateJobException(f"Required output not found: {output.name}")

    # check that the required output folders are present
    for output in processor_spec.outputFolders or []:
        if not any(x.name == output.name and x.isFolder for x in output_files_from_request):
            raise CreateJobException(f"Required output folder not found: {output.name}")

    # check that the input parameters are consistent with the spec
    for input_parameter in input_parameters:
        pp = next((x for x in processor_spec.parameters if x.name == input_parameter.name), None)
        if not pp:
            raise CreateJobException(f"Processor parameter not found: {input_parameter.name}")
        if not _parameter_value_is_consistent_with_type(input_parameter.value, pp.type):
            raise CreateJobException(f"Parameter value is not consistent with type: {input_parameter.name} {input_parameter.value} {pp.type}")

    # check that the parameters that do not have a default are present
    for pp in processor_spec.parameters:
        if pp.default is None and not pp.type.startswith('Optional['):
            input_parameter = next((x for x in input_parameters if x.name == pp.name), None)
            if not input_parameter:
                raise CreateJobException(f"Required parameter not found: {pp.name}")

def _parameter_value_is_consistent_with_type(value: Any, type: str):
    if type == 'str':
        return isinstance(value, str)
    elif type == 'int':
        return isinstance(value, int)
    elif type == 'float':
        return isinstance(value, float) or isinstance(value, int)
    elif type == 'bool':
        return isinstance(value, bool)
    elif type == 'List[str]':
        return isinstance(value, list) and all(isinstance(x, str) for x in value)
    elif type == 'List[int]':
        return isinstance(value, list) and all(isinstance(x, int) for x in value)
    elif type == 'List[float]':
        return isinstance(value, list) and all(isinstance(x, float) or isinstance(x, int) for x in value)
    elif type == 'List[bool]':
        return isinstance(value, list) and all(isinstance(x, bool) for x in value)
    elif type == 'Optional[int]':
        return isinstance(value, int) or value is None
    elif type == 'Optional[float]':
        return isinstance(value, float) or isinstance(value, int) or value is None
    else:
        raise Exception(f"Unknown type in _parameter_value_is_consistent_with_type: {type}")



import os
import json
from typing import Union
from ...common.dendro_types import DendroProject


def _get_project_role(project: DendroProject, user_id: Union[str, None]) -> str:
    ADMIN_USER_IDS_JSON = os.getenv('ADMIN_USER_IDS', '[]')
    ADMIN_USER_IDS = json.loads(ADMIN_USER_IDS_JSON)
    if user_id:
        if user_id in ADMIN_USER_IDS:
            return 'admin'
        if project.ownerId == user_id:
            return 'admin'
        user = next((x for x in project.users if x.userId == user_id), None)
        if user:
            return user.role
    if project.publiclyReadable:
        return 'viewer'
    else:
        return 'none'

def _project_has_user(project: DendroProject, user_id: Union[str, None]) -> bool:
    if not user_id:
        return False
    if project.ownerId == user_id:
        return True
    user = next((x for x in project.users if x.userId == user_id), None)
    if user:
        return True
    return False

def _check_user_can_read_project(project: DendroProject, user_id: Union[str, None]):
    if not _get_project_role(project, user_id) in ['admin', 'editor', 'viewer']:
        raise Exception('User does not have read permission for this project')

def _check_user_can_edit_project(project: DendroProject, user_id: Union[str, None]):
    if not _get_project_role(project, user_id) in ['admin', 'editor']:
        raise Exception('User does not have edit permission for this project')

def _check_user_is_project_admin(project: DendroProject, user_id: Union[str, None]):
    if not _get_project_role(project, user_id) == 'admin':
        raise Exception('User does not have admin permission for this project')



import os
import json
from typing import List, Optional
import time
from fastapi import APIRouter, Header
from .... import BaseModel
from ...core._create_random_id import _create_random_id
from ....common.dendro_types import DendroJob, DendroProject, DendroProjectUser
from ._authenticate_gui_request import _authenticate_gui_request
from ...core._get_project_role import _check_user_can_edit_project, _check_user_is_project_admin
from ...clients.db import fetch_project, insert_project, update_project, fetch_project_jobs, fetch_projects_for_user, fetch_all_projects, fetch_projects_with_tag
from ...services.gui.delete_project import delete_project as service_delete_project
from ..common import api_route_wrapper


router = APIRouter()

# get project
class GetProjectResponse(BaseModel):
    project: DendroProject
    success: bool

@router.get("/{project_id}")
@api_route_wrapper
async def get_project(project_id):
    project = await fetch_project(project_id, raise_on_not_found=True)
    assert project
    return GetProjectResponse(project=project, success=True)

# get projects
class GetProjectsResponse(BaseModel):
    projects: List[DendroProject]
    success: bool

@router.get("")
@api_route_wrapper
async def get_projects(github_access_token: str = Header(...), tag: Optional[str] = None):
    if tag is None:
        user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=False)
        # note: user may be None if they are not authenticated
        projects = await fetch_projects_for_user(user_id)
        return GetProjectsResponse(projects=projects, success=True)
    else:
        projects = await fetch_projects_with_tag(tag)
        return GetProjectsResponse(projects=projects, success=True)

# create project
class CreateProjectRequest(BaseModel):
    name: str

class CreateProjectResponse(BaseModel):
    projectId: str
    success: bool

@router.post("")
@api_route_wrapper
async def create_project(data: CreateProjectRequest, github_access_token: str = Header(...)) -> CreateProjectResponse:
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    # parse the request
    name = data.name

    project_id = _create_random_id(8)
    project = DendroProject(
        projectId=project_id,
        name=name,
        description='',
        ownerId=user_id,
        users=[],
        publiclyReadable=True,
        computeResourceId=None,
        tags=[],
        timestampCreated=time.time(),
        timestampModified=time.time()
    )

    await insert_project(project)

    return CreateProjectResponse(projectId=project_id, success=True)

# set project name
class SetProjectNameRequest(BaseModel):
    name: str

class SetProjectNameResponse(BaseModel):
    success: bool

@router.put("/{project_id}/name")
@api_route_wrapper
async def set_project_name(project_id, data: SetProjectNameRequest, github_access_token: str = Header(...)) -> SetProjectNameResponse:
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    project = await fetch_project(project_id, raise_on_not_found=True)
    assert project

    _check_user_can_edit_project(project, user_id)

    # parse the request
    name = data.name

    await update_project(project_id, update={
        'name': name,
        'timestampModified': time.time()
    })

    return SetProjectNameResponse(success=True)

# set project description
class SetProjectDescriptionRequest(BaseModel):
    description: str

class SetProjectDescriptionResponse(BaseModel):
    success: bool

@router.put("/{project_id}/description")
@api_route_wrapper
async def set_project_description(project_id, data: SetProjectDescriptionRequest, github_access_token: str = Header(...)) -> SetProjectDescriptionResponse:
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    project = await fetch_project(project_id, raise_on_not_found=True)
    assert project

    _check_user_can_edit_project(project, user_id)

    # parse the request
    description = data.description

    await update_project(project_id, update={
        'description': description,
        'timestampModified': time.time()
    })

    return SetProjectDescriptionResponse(success=True)

# set project tags
class SetProjectTagsRequest(BaseModel):
    tags: List[str]

class SetProjectTagsResponse(BaseModel):
    success: bool

@router.put("/{project_id}/tags")
@api_route_wrapper
async def set_project_tags(project_id, data: SetProjectTagsRequest, github_access_token: str = Header(...)) -> SetProjectTagsResponse:
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    project = await fetch_project(project_id, raise_on_not_found=True)
    assert project

    _check_user_can_edit_project(project, user_id)

    # parse the request
    tags = data.tags

    await update_project(project_id, update={
        'tags': tags,
        'timestampModified': time.time()
    })

    return SetProjectTagsResponse(success=True)

# delete project
class DeleteProjectResponse(BaseModel):
    success: bool

@router.delete("/{project_id}")
@api_route_wrapper
async def delete_project(project_id, github_access_token: str = Header(...)) -> DeleteProjectResponse:
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    project = await fetch_project(project_id, raise_on_not_found=True)
    assert project

    _check_user_is_project_admin(project, user_id)

    await service_delete_project(project)

    return DeleteProjectResponse(success=True)

# get jobs
class GetJobsResponse(BaseModel):
    jobs: List[DendroJob]
    success: bool

@router.get("/{project_id}/jobs")
@api_route_wrapper
async def get_jobs(project_id):
    jobs = await fetch_project_jobs(project_id, include_private_keys=False)
    return GetJobsResponse(jobs=jobs, success=True)

# set project publicly readable
class SetProjectPubliclyReadableRequest(BaseModel):
    publiclyReadable: bool

class SetProjectPubliclyReadableResponse(BaseModel):
    success: bool

@router.put("/{project_id}/publicly_readable")
@api_route_wrapper
async def set_project_public(project_id, data: SetProjectPubliclyReadableRequest, github_access_token: str = Header(...)):
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    project = await fetch_project(project_id, raise_on_not_found=True)
    assert project

    _check_user_is_project_admin(project, user_id)

    # parse the request
    publicly_readable = data.publiclyReadable

    await update_project(project_id, update={
        'publiclyReadable': publicly_readable,
        'timestampModified': time.time()
    })

    return SetProjectPubliclyReadableResponse(success=True)

# set project compute resource id
class SetProjectComputeResourceIdRequest(BaseModel):
    computeResourceId: str

class SetProjectComputeResourceIdResponse(BaseModel):
    success: bool

@router.put("/{project_id}/compute_resource_id")
@api_route_wrapper
async def set_project_compute_resource_id(project_id, data: SetProjectComputeResourceIdRequest, github_access_token: str = Header(...)):
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    project = await fetch_project(project_id, raise_on_not_found=True)
    assert project

    _check_user_is_project_admin(project, user_id)

    # parse the request
    compute_resource_id = data.computeResourceId

    await update_project(project_id, update={
        'computeResourceId': compute_resource_id,
        'timestampModified': time.time()
    })

    return SetProjectComputeResourceIdResponse(success=True)

class SetProjectAnalysisSourceUrlRequest(BaseModel):
    analysisSourceUrl: str

class SetProjectAnalysisSourceUrlResponse(BaseModel):
    success: bool

@router.put("/{project_id}/analysis_source_url")
@api_route_wrapper
async def set_project_analysis_source_url(project_id, data: SetProjectAnalysisSourceUrlRequest, github_access_token: str = Header(...)):
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    project = await fetch_project(project_id, raise_on_not_found=True)
    assert project

    _check_user_can_edit_project(project, user_id)

    # parse the request
    analysis_source_url = data.analysisSourceUrl

    await update_project(project_id, update={
        'analysisSourceUrl': analysis_source_url,
        'timestampModified': time.time()
    })

    return SetProjectAnalysisSourceUrlResponse(success=True)

# set project users
class SetProjectUsersRequest(BaseModel):
    users: List[DendroProjectUser]

class SetProjectUsersResponse(BaseModel):
    success: bool

@router.put("/{project_id}/users")
@api_route_wrapper
async def set_project_users(project_id, data: SetProjectUsersRequest, github_access_token: str = Header(...)):
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    project = await fetch_project(project_id, raise_on_not_found=True)
    assert project

    _check_user_is_project_admin(project, user_id)

    # parse the request
    # support both pydantic v1 and v2
    users = [_model_dump(x) for x in data.users]

    await update_project(project_id, update={
        'users': users,
        'timestampModified': time.time()
    })

    return SetProjectUsersResponse(success=True)

# Admin get all projects
class AdminGetAllProjectsResponse(BaseModel):
    projects: List[DendroProject]
    success: bool

@router.get("/admin/get_all_projects")
@api_route_wrapper
async def admin_get_all_projects(github_access_token: str = Header(...)):
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    ADMIN_USER_IDS_JSON = os.getenv('ADMIN_USER_IDS', '[]')
    ADMIN_USER_IDS = json.loads(ADMIN_USER_IDS_JSON)

    if user_id not in ADMIN_USER_IDS:
        raise Exception('User is not admin')

    projects = await fetch_all_projects()
    return AdminGetAllProjectsResponse(projects=projects, success=True)

def _model_dump(model, exclude_none=False):
    # handle both pydantic v1 and v2
    if hasattr(model, 'model_dump'):
        return model.model_dump(exclude_none=exclude_none)
    else:
        return model.dict(exclude_none=exclude_none)
